@@ -607,16 +607,24 @@ hwloc_levelzero_discover(struct hwloc_backend *backend, struct hwloc_disc_status
   int sysman_maybe_missing = 0; /* 1 if ZES_ENABLE_SYSMAN=1 was NOT set early, 2 if ZES_ENABLE_SYSMAN=0 */
   char *env;
 
-  hwloc__levelzero_osdev_array_init(&oarray);
-
-  hwloc__levelzero_ports_init(&hports);
-
   assert(dstatus->phase == HWLOC_DISC_PHASE_IO);
 
   hwloc_topology_get_type_filter(topology, HWLOC_OBJ_OS_DEVICE, &filter);
   if (filter == HWLOC_TYPE_FILTER_KEEP_NONE)
     return 0;
 
+  hwloc__levelzero_osdev_array_init(&oarray);
+
+  hwloc__levelzero_ports_init(&hports);
+
+#ifdef HWLOC_HAVE_ZESINIT
+  res = zesInit(0);
+  if (res != ZE_RESULT_SUCCESS) {
+    hwloc_debug("hwloc/levelzero: Failed to initialize LevelZero Sysman in zesInit(): 0x%x\n", (unsigned)res);
+    hwloc_debug("hwloc/levelzero: Continuing. Hopefully ZES_ENABLE_SYSMAN=1\n");
+  }
+#endif /* HWLOC_HAVE_ZESINIT */
+
   /* Tell L0 to create sysman devices.
    * If somebody already initialized L0 without Sysman,
    * zesDeviceGetProperties() will fail and warn in hwloc__levelzero_properties_get().
@@ -679,6 +687,15 @@ hwloc_levelzero_discover(struct hwloc_backend *backend, struct hwloc_disc_status
       continue;
     }
 
+#if 0
+    /* No interesting info attr to get from driver properties for now.
+     * 2022/12/09: Driver UUID is driver version (major>>24|minor>>16|build) | a-uuid-timestamp>>32
+     *  hence it's not stable across multiple runs
+     */
+    ze_driver_properties_t drprop;
+    res = zeDriverGetProperties(drh[i], &drprop);
+#endif
+
     for(j=0; j<nbdevices; j++) {
       zes_device_handle_t sdvh = dvh[j];
       zes_device_handle_t *subh = NULL;
