@@ -1,6 +1,6 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2021 Inria.  All rights reserved.
+ * Copyright © 2009-2023 Inria.  All rights reserved.
  * Copyright © 2009-2010, 2012, 2015 Université Bordeaux
  * Copyright © 2011 Cisco Systems, Inc.  All rights reserved.
  * Copyright © 2020 Hewlett Packard Enterprise.  All rights reserved.
@@ -13,6 +13,7 @@
 #include "private/autogen/config.h"
 #include "hwloc.h"
 #include "misc.h"
+#include "hwloc-calc.h"
 
 enum lstopo_drawing_e {
   LSTOPO_DRAWING_PREPARE,
@@ -100,7 +101,7 @@ struct lstopo_output {
   int show_memattrs_only;
   int show_cpukinds_only;
   int show_windows_processor_groups_only;
-  hwloc_obj_type_t show_only;
+  struct hwloc_calc_level show_only;
   int show_cpuset;
   int show_taskset;
   int transform_distances;
@@ -111,6 +112,7 @@ struct lstopo_output {
   unsigned int gridsize, fontsize, linespacing, thickness;
   float text_xscale;
   enum lstopo_orient_e force_orient[HWLOC_OBJ_TYPE_MAX]; /* orientation of children within an object of the given type */
+  enum lstopo_orient_e above_force_orient;
   enum lstopo_orient_e right_force_orient;
   enum lstopo_orient_e below_force_orient;
   int show_indexes[HWLOC_OBJ_TYPE_MAX]; /* enabled by global toggle index_type */
@@ -122,6 +124,8 @@ struct lstopo_output {
   int show_disallowed;
   int show_process_color;
   int show_cpukinds;
+  char *logical_index_prefix;
+  char *os_index_prefix;
   int factorize_enabled; /* global toggle for interactive keyboard shortcuts */
   unsigned factorize_min[HWLOC_OBJ_TYPE_MAX]; /* minimum number of object before factorizing (parent->arity must be strictly higher) */
 #define FACTORIZE_MIN_DEFAULT 4
